"""
``LiveEnhancedPbpItem`` is the base class for all live data enhanced pbp event types
"""
from collections import defaultdict

from pbpstats.resources.enhanced_pbp.enhanced_pbp_item import EnhancedPbpItem
from pbpstats.resources.enhanced_pbp import StartOfPeriod

KEY_ATTR_MAPPER = {
    "period": "period",
    "actionNumber": "event_num",
    "clock": "clock",
    "description": "description",
    "xLegacy": "locX",
    "yLegacy": "locY",
    "actionType": "action_type",
    "subType": "sub_type",
    "descriptor": "descriptor",
    "shotResult": "shot_result",
    "assistPersonId": "player2_id",
    "blockPersonId": "player3_id",
    "stealPersonId": "player3_id",
    "foulDrawnPersonId": "player3_id",
    "shotActionNumber": "shot_action_number",
    "qualifiers": "qualifiers",
    "teamId": "team_id",
    "personId": "player1_id",
    "scoreHome": "home_score",
    "scoreAway": "away_score",
    "possession": "offense_team_id",
    "orderNumber": "order",
}


class LiveEnhancedPbpItem(EnhancedPbpItem):
    """
    Base class for enhanced pbp events from live data

    :param dict item: dict with event data
    :param int period: period in which event occurs
    :param str game_id: NBA Stats Game Id
    """

    def __init__(self, item, game_id):
        self.game_id = game_id
        for key, value in KEY_ATTR_MAPPER.items():
            attr_value = item.get(key)
            if attr_value is not None and attr_value != "":
                try:
                    # convert string values that should be ints
                    setattr(self, value, int(attr_value))
                except (ValueError, TypeError):
                    setattr(self, value, attr_value)
            if key == "description" and attr_value is None:
                setattr(self, value, "")
        self.player_game_fouls = defaultdict(int)
        self.possession_changing_override = False
        self.non_possession_changing_override = False
        self.score = defaultdict(int)

    @property
    def data(self):
        """
        returns event as a dict
        """
        return self.__dict__

    @property
    def event_stats(self):
        """
        returns list of dicts with all stats for event
        """
        return self.base_stats

    def get_offense_team_id(self):
        """
        returns team id for team on offense for event
        """
        return self.offense_team_id

    @property
    def is_possession_ending_event(self):
        """
        returns True if event ends a possession, False otherwise
        """
        if self.next_event is None:
            return True
        elif self.next_event is not None:
            oftid_changed = self.offense_team_id != self.next_event.offense_team_id
            # offense_team_id is 0 for first event of first quarter and overtimes - don't count it
            # sometimes there is an event before period start event that breaks things so check that too
            oftid_is_0 = (
                self.offense_team_id == 0 or self.next_event.offense_team_id == 0
            )
            start_of_period = isinstance(self, StartOfPeriod) or isinstance(
                self.next_event, StartOfPeriod
            )
            if oftid_changed and not oftid_is_0 and not start_of_period:
                return True
        return False

    @property
    def seconds_remaining(self):
        """
        returns seconds remaining in period as a ``float``
        """
        stripped = self.clock.replace("PT", "").replace("M", ":").replace("S", "")
        split = stripped.split(":")
        return float(split[0]) * 60 + float(split[1])

    @property
    def stripped_sub_type(self):
        return self.sub_type.replace("-", "").replace(" ", "")

    @property
    def stripped_descriptor(self):
        return self.descriptor.replace("-", "").replace(" ", "")



OVERRIDES = {
    "PLAYER_NAME": "name",
}


class StatsNbaBoxscoreItem(object):
    """
    Class for boxscore items from stats.nba.com

    :param dict item: dict with boxscore stats from response
    """

    def __init__(self, item):
        for key, value in item.items():
            setattr(self, OVERRIDES.get(key, key.lower()), value)

    @property
    def data(self):
        """
        returns boxscore data dict
        """
        return self.__dict__



from pbpstats.resources.enhanced_pbp.live.enhanced_pbp_item import LiveEnhancedPbpItem
from pbpstats.resources.enhanced_pbp import FreeThrow


class LiveFreeThrow(FreeThrow, LiveEnhancedPbpItem):
    """
    Class for free throw events
    """

    action_type = "freethrow"

    def __init__(self, *args):
        super().__init__(*args)

    @property
    def is_made(self):
        """
        returns True if shot was made, False otherwise
        """
        return self.shot_result == "Made"

    @property
    def is_ft_1_of_1(self):
        return self.sub_type == "1 of 1" and not hasattr(self, "descriptor")

    @property
    def is_ft_1_of_2(self):
        return self.sub_type == "1 of 2"

    @property
    def is_ft_2_of_2(self):
        return self.sub_type == "2 of 2"

    @property
    def is_ft_1_of_3(self):
        return self.sub_type == "1 of 3"

    @property
    def is_ft_2_of_3(self):
        return self.sub_type == "2 of 3"

    @property
    def is_ft_3_of_3(self):
        return self.sub_type == "3 of 3"

    @property
    def is_technical_ft(self):
        return hasattr(self, "descriptor") and self.descriptor == "technical"

    @property
    def is_ft_1pt(self):
        """
        returns True if free throw is a 1 point free throw, False otherwise
        Only used in g-league, starting in 2019-20 season
        """
        # TODO: add this if/when this is available for gleague
        return False

    @property
    def is_ft_2pt(self):
        """
        returns True if free throw is a 2 point free throw, False otherwise
        Only used in g-league, starting in 2019-20 season
        """
        # TODO: add this if/when this is available for gleague
        return False

    @property
    def is_ft_3pt(self):
        """
        returns True if free throw is a 3 point free throw, False otherwise
        Only used in g-league, starting in 2019-20 season
        """
        # TODO: add this if/when this is available for gleague
        return False

    @property
    def is_away_from_play_ft(self):
        """
        returns True if free throw is from an away from the play foul, False otherwise.
        """
        return (
            hasattr(self, "descriptor") and self.stripped_descriptor == "awayfromplay"
        )

    @property
    def is_flagrant_ft(self):
        """
        returns True if free throw is from a flagrant foul, False otherwise.
        """
        return hasattr(self, "descriptor") and self.stripped_descriptor == "flagrant"



import json
import os

from pbpstats import (
    NBA_STRING,
    G_LEAGUE_STRING,
    WNBA_STRING,
    NBA_GAME_ID_PREFIX,
    G_LEAGUE_GAME_ID_PREFIX,
    WNBA_GAME_ID_PREFIX,
)
from pbpstats.data_loader.stats_nba.web_loader import StatsNbaWebLoader


class StatsNbaScoreboardWebLoader(StatsNbaWebLoader):
    """
    A ``StatsNbaScoreboardWebLoader`` object should be instantiated and passed into ``StatsNbaScoreboardLoader`` when loading data directly from the NBA Stats API

    :param str file_directory: (optional, use it if you want to store the response data on disk)
        Directory in which data should be either stored.
        The specific file location will be `stats_<league>_<date>.json` in the `/schedule` subdirectory.
        If not provided response data will not be saved on disk.
    """
    def __init__(self, file_directory=None):
        self.file_directory = file_directory

    def load_data(self, date, league_string):
        self.date = date
        self.league_string = league_string
        league_url_part = (
            f"{G_LEAGUE_STRING}.{NBA_STRING}"
            if self.league_string == G_LEAGUE_STRING
            else self.league_string
        )
        self.base_url = f"https://stats.{league_url_part}.com/stats/scoreboardV2"
        self.parameters = {
            "DayOffset": 0,
            "LeagueID": self.league_id,
            "gameDate": self.date,
        }
        return self._load_request_data()

    def _save_data_to_file(self):
        if self.file_directory is not None and os.path.isdir(self.file_directory):
            file_path = f'{self.file_directory}/schedule/stats_{self.league_string}_{self.date.replace("/", "_")}.json'
            with open(file_path, "w") as outfile:
                json.dump(self.source_data, outfile)

    @property
    def league_id(self):
        """
        Returns League Id for league.

        00 for nba, 10 for wnba, 20 for g-league
        """
        if self.league_string == NBA_STRING:
            return NBA_GAME_ID_PREFIX
        elif self.league_string == WNBA_STRING:
            return WNBA_GAME_ID_PREFIX
        elif self.league_string == G_LEAGUE_STRING:
            return G_LEAGUE_GAME_ID_PREFIX



from pbpstats.resources.enhanced_pbp.live.enhanced_pbp_item import LiveEnhancedPbpItem
from pbpstats.resources.enhanced_pbp import Substitution


class LiveSubstitution(Substitution, LiveEnhancedPbpItem):
    """
    Class for Substitution events
    """

    action_type = "substitution"

    def __init__(self, *args):
        super().__init__(*args)

    @property
    def incoming_player_id(self):
        """
        returns player id of player coming in to the game
        """
        if self.sub_type == "out":
            return None
        return self.player1_id

    @property
    def outgoing_player_id(self):
        """
        returns player id of player coming in to the game
        """
        if self.sub_type == "in":
            return None
        return self.player1_id

    @property
    def current_players(self):
        """
        returns dict with list of player ids for each team
        with players on the floor following the sub
        """
        players = {}
        for team_id, team_players in self.previous_event.current_players.items():
            players[team_id] = [player_id for player_id in team_players]
        if self.player1_id is not None:
            if self.sub_type == "in":
                players[self.team_id].append(self.player1_id)
            elif self.sub_type == "out" and self.player1_id in players[self.team_id]:
                players[self.team_id].remove(self.player1_id)
            players[self.team_id] = list(set(players[self.team_id]))
        return players



from pbpstats.resources.enhanced_pbp.live.enhanced_pbp_item import LiveEnhancedPbpItem
from pbpstats.resources.enhanced_pbp import Ejection


class LiveEjection(Ejection, LiveEnhancedPbpItem):
    """
    Class for Ejection events
    """

    action_type = "ejection"

    def __init__(self, *args):
        super().__init__(*args)



from pbpstats import (
    NBA_STRING,
    G_LEAGUE_STRING,
    WNBA_STRING,
    PLAYOFFS_STRING,
    REGULAR_SEASON_STRING,
    PLAY_IN_STRING,
    NBA_GAME_ID_PREFIX,
    G_LEAGUE_GAME_ID_PREFIX,
    WNBA_GAME_ID_PREFIX,
)


class StatsNbaLoaderBase(object):
    """
    Base Class for all stats.nba.com data loaders

    This class should not be instantiated directly
    """

    def make_list_of_dicts(self, results_set_index=0):
        """
        Creates list of dicts from source data

        :param int results_set_index: Index results are in. Default is 0
        :returns: list of dicts with data for results
        """
        headers = self.source_data["resultSets"][results_set_index]["headers"]
        rows = self.source_data["resultSets"][results_set_index]["rowSet"]
        deduped_rows = self.dedupe_events_row_set(rows)
        return [dict(zip(headers, row)) for row in deduped_rows]

    @staticmethod
    def dedupe_events_row_set(events_row_set):
        """
        Dedupes list of results while preserving order

        Used to dedupe events rowSets pbp response because some games have duplicate events

        :param list events_row_set: List of results from API Response
        :returns: deduped list of results
        """
        deduped_events_row_set = []
        for sublist in events_row_set:
            if sublist not in deduped_events_row_set:
                deduped_events_row_set.append(sublist)
        return deduped_events_row_set

    @property
    def data(self):
        """
        returns data from response JSON as a list of dicts
        """
        return self.make_list_of_dicts()

    @property
    def league(self):
        """
        Returns League for game id.

        First 2 in game id represent league - 00 for nba, 10 for wnba, 20 for g-league
        """
        if self.game_id[0:2] == NBA_GAME_ID_PREFIX:
            return NBA_STRING
        elif self.game_id[0:2] == G_LEAGUE_GAME_ID_PREFIX:
            return G_LEAGUE_STRING
        elif self.game_id[0:2] == WNBA_GAME_ID_PREFIX:
            return WNBA_STRING

    @property
    def season(self):
        """
        Returns season for game id

        4th and 5th characters in game id represent season year
        ex. for 2016-17 season 4th and 5th characters would be 16 and season should return 2016-17
        For WNBA just returns season year
        """
        digit4 = int(self.game_id[3])
        digit5 = int(self.game_id[4])
        if digit4 == 9:
            if digit5 == 9:
                return "1999" if self.league == WNBA_STRING else "1999-00"
            else:
                return (
                    f"19{digit4}{digit5}"
                    if self.league == WNBA_STRING
                    else f"19{digit4}{digit5}-{digit4}{digit5 + 1}"
                )
        elif digit5 == 9:
            return (
                f"20{digit4}{digit5}"
                if self.league == WNBA_STRING
                else f"20{digit4}{digit5}-{digit4 + 1}0"
            )
        else:
            return (
                f"20{digit4}{digit5}"
                if self.league == WNBA_STRING
                else f"20{digit4}{digit5}-{digit4}{digit5 + 1}"
            )

    @property
    def season_type(self):
        """
        Returns season type for game id

        3rd character in game id represent season type - 2 for reg season, 4 for playoffs, 5 for play in
        """
        if self.game_id[2] == "4":
            return PLAYOFFS_STRING
        elif self.game_id[2] == "2":
            return REGULAR_SEASON_STRING
        elif self.game_id[2] == "5":
            return PLAY_IN_STRING

    @property
    def league_id(self):
        """
        Returns League Id for league.

        00 for nba, 10 for wnba, 20 for g-league
        """
        if self.league_string == NBA_STRING:
            return NBA_GAME_ID_PREFIX
        elif self.league_string == WNBA_STRING:
            return WNBA_GAME_ID_PREFIX
        elif self.league_string == G_LEAGUE_STRING:
            return G_LEAGUE_GAME_ID_PREFIX


"""
``DataNbaBoxscoreLoader`` loads boxscore data for a game and creates :obj:`~pbpstats.resources.boxscore.data_nba_boxscore_item.DataNbaBoxscoreItem` objects for each player and team

The following code will load boxscore data for game id "0021900001" from a file located in a subdirectory of the /data directory

.. code-block:: python

    from pbpstats.data_loader import DataNbaBoxscoreFileLoader, DataNbaBoxscoreLoader

    source_loader = DataNbaBoxscoreFileLoader("/data")
    boxscore_loader = DataNbaBoxscoreLoader("0021900001", source_loader)
    print(boxscore_loader.items[0].data) # prints dict with a player's boxscore data for game
"""
from pbpstats.data_loader.data_nba.base import DataNbaLoaderBase
from pbpstats.resources.boxscore.data_nba_boxscore_item import DataNbaBoxscoreItem


class DataNbaBoxscoreLoader(DataNbaLoaderBase):
    """
    Loads data.nba.com source boxscore data for game.
    Team/Player data is stored in items attribute as :obj:`~pbpstats.resources.boxscore.data_nba_boxscore_item.DataNbaBoxscoreItem` objects

    :param str game_id: NBA Stats Game Id
    :param source_loader: :obj:`~pbpstats.data_loader.data_nba.boxscore.file.DataNbaBoxscoreFileLoader` or :obj:`~pbpstats.data_loader.data_nba.boxscore.web.DataNbaBoxscoreWebLoader` object
    """

    data_provider = "data_nba"
    resource = "Boxscore"
    parent_object = "Game"

    def __init__(self, game_id, source_loader):
        self.game_id = game_id
        self.source_data = source_loader.load_data(self.game_id)
        self._make_boxscore_items()

    def _make_boxscore_items(self):
        """
        makes :obj:`~pbpstats.resources.boxscore.DataNbaBoxscoreItem` items for each player/team
        """
        home = self.data["hls"]
        away = self.data["vls"]
        self.items = [
            DataNbaBoxscoreItem(item, team_id=away["tid"], team_abbreviation=away["ta"])
            for item in away["pstsg"]
        ]
        self.items += [
            DataNbaBoxscoreItem(item, team_id=home["tid"], team_abbreviation=home["ta"])
            for item in home["pstsg"]
        ]
        self.items.append(
            DataNbaBoxscoreItem(
                away["tstsg"], team_id=away["tid"], team_abbreviation=away["ta"]
            )
        )
        self.items.append(
            DataNbaBoxscoreItem(
                home["tstsg"], team_id=home["tid"], team_abbreviation=home["ta"]
            )
        )

    @property
    def data(self):
        """
        returns raw JSON response data
        """
        return self.source_data["g"]



import requests

from pbpstats.data_loader.live.base import LiveLoaderBase


class LiveWebLoader(LiveLoaderBase):
    """
    Base class for loading data from live data API request.

    All live data data loader classes should inherit from this class.

    This class should not be instantiated directly.
    """

    def _load_request_data(self):
        response = requests.get(self.url)
        if response.status_code == 200:
            self.source_data = response.json()
            self._save_data_to_file()
            return self.source_data
        else:
            response.raise_for_status()



from pbpstats.resources.enhanced_pbp.data_nba.enhanced_pbp_item import (
    DataEnhancedPbpItem,
)
from pbpstats.resources.enhanced_pbp import Substitution


class DataSubstitution(Substitution, DataEnhancedPbpItem):
    """
    Class for Substitution events
    """

    event_type = 8

    def __init__(self, *args):
        super().__init__(*args)

    @property
    def outgoing_player_id(self):
        """
        returns player id of player going out of the game
        """
        return self.player1_id

    @property
    def incoming_player_id(self):
        """
        returns player id of player coming in to the game
        """
        return self.player2_id



"""
The ``Pbp`` class has some basic properties for handling pbp data
"""
from pbpstats.resources.base import Base


class Pbp(Base):
    """
    Class for pbp items

    :param list items: list of either
        :obj:`~pbpstats.resources.pbp.stats_nba_pbp_item.StatsNbaPbpItem` or
        :obj:`~pbpstats.resources.pbp.data_nba_pbp_item.DataNbaPbpItem` items,
        typically from a pbp data loader
    """

    def __init__(self, items):
        self.items = items

    @property
    def data(self):
        """
        returns list of dicts with each event
        """
        return [item.data for item in self.items]



from pbpstats.data_loader.abs_data_loader import check_file_directory
from pbpstats.data_loader.stats_nba.file_loader import StatsNbaFileLoader



class StatsNbaGameFinderFileLoader(StatsNbaFileLoader):
    """
    A ``StatsNbaGameFinderFileLoader`` object should be instantiated and passed into ``StatsNbaGameFinderLoader`` when loading data from file

    :param str file_directory:
        Directory in which data should be loaded from.
        The specific file location will be `stats_<league>_<season>_<season_type>.json` in the `/schedule` subdirectory.
    """
    def __init__(self, file_directory):
        self.file_directory = file_directory

    @check_file_directory
    def load_data(self, league, season, season_type):
        self.league_string = league
        self.season_string = season
        self.season_type_string = season_type
        self.file_path = f'{self.file_directory}/schedule/stats_{self.league_string}_{self.season_string.replace("-", "_")}_{self.season_type_string.replace(" ", "_")}.json'
        return self._load_data_from_file()



from pbpstats.resources.enhanced_pbp.live.enhanced_pbp_item import LiveEnhancedPbpItem
from pbpstats.resources.enhanced_pbp import Rebound


class LiveRebound(Rebound, LiveEnhancedPbpItem):
    """
    Class for rebound events
    """

    action_type = "rebound"

    def __init__(self, *args):
        super().__init__(*args)

    @property
    def missed_shot(self):
        """
        returns :obj:`~pbpstats.resources.enhanced_pbp.field_goal.FieldGoal` or
        :obj:`~pbpstats.resources.enhanced_pbp.free_throw.FreeThrow` object
        for shot that was missed

        :raises: :obj:`~pbpstats.resources.enhanced_pbp.rebound.EventOrderError`:
            If rebound event is not immediately following a missed shot event.
        """
        prev_event = self.previous_event
        while prev_event is not None:
            if prev_event.event_num == self.shot_action_number:
                return prev_event
            prev_event = prev_event.previous_event

        raise Rebound.EventOrderError(
            f"previous event: {self.previous_event} is not a missed free throw or field goal"
        )

    @property
    def oreb(self):
        """
        returns True if rebound is an offensive rebound, False otherwise
        """
        return self.sub_type == "offensive"

    @property
    def is_placeholder(self):
        """
        returns True if rebound is a placeholder event, False otherwise.

        These are team rebounds on for example missed FT 1 of 2
        """
        if hasattr(self, "qualifiers") and "deadball" in self.qualifiers:
            return True
        if (
            hasattr(self.missed_shot, "is_flagrant_ft")
            and self.missed_shot.is_flagrant_ft
        ):
            return True
        return False



KEY_ATTR_MAPPER = {
    "gid": "game_id",
    "gdte": "date",
    "stt": "status",
}


class DataNbaGameItem(object):
    """
    Class for game data from data.nba.com

    :param dict item: dict with game data
    """

    def __init__(self, item):
        for key, value in KEY_ATTR_MAPPER.items():
            if item.get(key) is not None:
                setattr(self, value, item.get(key))
        self.home_team_id = item["h"]["tid"]
        self.home_team_abbreviation = item["h"]["ta"]
        self.home_score = item["h"]["s"]
        self.away_team_id = item["v"]["tid"]
        self.away_team_abbreviation = item["v"]["ta"]
        self.away_score = item["v"]["s"]

    @property
    def data(self):
        """
        returns game dict
        """
        return self.__dict__

    @property
    def is_final(self):
        """
        returns True if game is final, False otherwise
        """
        return self.status == "Final"



"""
``LiveEnhancedPbpLoader`` loads pbp data for a game and creates :obj:`~pbpstats.resources.enhanced_pbp.enhanced_pbp_item.EnhancedPbpItem` objects for each event

Enhanced data for each event includes current players on floor, score, fouls to give and number of fouls committed by each player,
plus additional data depending on event type

The following code will load pbp data for game id "0021900001" from a file located in a subdirectory of the /data directory

.. code-block:: python

    from pbpstats.data_loader import LiveEnhancedPbpFileLoader, LiveEnhancedPbpLoader

    source_loader = LiveEnhancedPbpFileLoader("/data")
    pbp_loader = LiveEnhancedPbpLoader("0021900001", source_loader)
    print(pbp_loader.items[0].data)  # prints dict with the first event of the game
"""
from pbpstats.data_loader.live.pbp.loader import LivePbpLoader
from pbpstats.data_loader.nba_enhanced_pbp_loader import NbaEnhancedPbpLoader
from pbpstats.resources.enhanced_pbp.live.enhanced_pbp_factory import (
    LiveEnhancedPbpFactory,
)
from pbpstats.resources.enhanced_pbp import Rebound


class LiveEnhancedPbpLoader(LivePbpLoader, NbaEnhancedPbpLoader):
    """
    Loads data.nba.com source enhanced pbp data for game.
    Events are stored in items attribute as :obj:`~pbpstats.resources.enhanced_pbp.enhanced_pbp_item.EnhancedPbpItem` objects

    :param str game_id: NBA Stats Game Id
    :param source_loader: :obj:`~pbpstats.data_loader.live.enhanced_pbp.file.LiveEnhancedPbpFileLoader` or :obj:`~pbpstats.data_loader.live.enhanced_pbp.web.LiveEnhancedPbpWebLoader` object
    """

    data_provider = "live"
    resource = "EnhancedPbp"
    parent_object = "Game"

    def __init__(self, game_id, source_loader):
        self.file_directory = source_loader.file_directory
        super().__init__(game_id, source_loader)

    def _make_pbp_items(self):
        factory = LiveEnhancedPbpFactory()
        self.items = [
            factory.get_event_class(event["actionType"], event.get("subType", ""))(
                event, self.game_id
            )
            for event in self.data
        ]
        self._add_extra_attrs_to_all_events()
        self._change_team_id_on_drebs()

    def _change_team_id_on_drebs(self):
        """
        live pbp changes possession on dreb, swap team to be consistent with other sources
        """
        for event in self.items:
            if isinstance(event, Rebound):
                if event.is_real_rebound and not event.oreb:
                    event.offense_team_id = event.previous_event.offense_team_id



"""
``StatsNbaPbpLoader`` loads pbp data for a game and
creates :obj:`~pbpstats.resources.pbp.stats_nba_pbp_item.StatsNbaPbpItem` objects for each event

The following code will load pbp data for game id "0021900001" from a file
located in a subdirectory of the /data directory

.. code-block:: python

    from pbpstats.data_loader import StatsNbaPbpFileLoader, StatsNbaPbpLoader

    source_loader = StatsNbaPbpFileLoader("/data")
    pbp_loader = StatsNbaPbpLoader("0021900001", source_loader)
    print(pbp_loader.items[0].data)  # prints dict with the first event of the game
"""
from pbpstats.data_loader.stats_nba.base import StatsNbaLoaderBase
from pbpstats.resources.pbp.stats_nba_pbp_item import StatsNbaPbpItem


class StatsNbaPbpLoader(StatsNbaLoaderBase):
    """
    Loads stats.nba.com source pbp data for game.
    Events are stored in items attribute
    as :obj:`~pbpstats.resources.pbp.stats_nba_pbp_item.StatsNbaPbpItem` objects

    :param str game_id: NBA Stats Game Id
    :param source_loader: :obj:`~pbpstats.data_loader.stats_nba.pbp.file.StatsNbaPbpFileLoader` or :obj:`~pbpstats.data_loader.stats_nba.pbp.file.StatsNbaPbpWebLoader` object
    """

    data_provider = "stats_nba"
    resource = "Pbp"
    parent_object = "Game"

    def __init__(self, game_id, source_loader):
        self.game_id = game_id
        self.source_data = source_loader.load_data(self.game_id)
        self.file_directory = source_loader.file_directory
        self._make_pbp_items()

    def _make_pbp_items(self):
        self.items = [StatsNbaPbpItem(item, i) for i, item in enumerate(self.data)]



class StatsNbaPbpItem(object):
    """
    Class for pbp events from stats.nba.com

    :param dict event: dict with event data
    :param int order: sequential order in which event occurs
    """

    def __init__(self, event, order):
        for key, value in event.items():
            setattr(self, key.lower(), value)
        self.order = order

    @property
    def data(self):
        """
        returns pbp event dict
        """
        return self.__dict__



import json
from pathlib import Path

from pbpstats.data_loader.abs_data_loader import check_file_directory
from pbpstats.data_loader.stats_nba.file_loader import StatsNbaFileLoader


class StatsNbaShotsFileLoader(StatsNbaFileLoader):
    """
    A ``StatsNbaShotsFileLoader`` object should be instantiated and passed into ``StatsNbaShotsLoader`` when loading data from file

    :param str file_directory:
        Directory in which data should be loaded from.
        The specific file location will be `stats_home_shots_<game_id>.json`
        and `stats_away_shots_<game_id>.json` in the `/game_details` subdirectory.
    """
    def __init__(self, file_directory):
        self.file_directory = file_directory

    @check_file_directory
    def load_data(self, game_id):
        self.game_id = game_id
        self.home_file_path = (
            f"{self.file_directory }/game_details/stats_home_shots_{self.game_id}.json"
        )
        self.away_file_path = (
            f"{self.file_directory }/game_details/stats_away_shots_{self.game_id}.json"
        )

        home_data_file = Path(self.home_file_path)
        if not home_data_file.is_file():
            raise Exception(f"{self.home_file_path} does not exist")
        with open(self.home_file_path) as json_data:
            self.home_source_data = json.load(json_data)

        away_data_file = Path(self.away_file_path)
        if not away_data_file.is_file():
            raise Exception(f"{self.away_file_path} does not exist")
        with open(self.away_file_path) as json_data:
            self.away_source_data = json.load(json_data)

        return self.home_source_data, self.away_source_data



from pbpstats.data_loader.data_nba.enhanced_pbp.file import DataNbaEnhancedPbpFileLoader


class DataNbaPossessionFileLoader(object):
    """
    A ``DataNbaPossessionFileLoader`` object should be instantiated and passed into ``DataNbaPossessionLoader`` when loading data from file

    :param str file_directory:
        Directory in which data should be loaded from.
        The specific file location will be `data_<game_id>.json` in the `/pbp` subdirectory.
    """
    def __init__(self, file_directory=None):
        self.file_directory = file_directory
        self.enhanced_pbp_source_loader = DataNbaEnhancedPbpFileLoader(file_directory)


"""
``DataNbaPbpLoader`` loads pbp data for a game and creates :obj:`~pbpstats.resources.pbp.data_nba_pbp_item.DataNbaPbpItem` objects for each event

The following code will load pbp data for game id "0021900001" from a file located in a subdirectory of the /data directory

.. code-block:: python

    from pbpstats.data_loader import DataNbaPbpFileLoader, DataNbaPbpLoader

    source_loader = DataNbaPbpFileLoader("/data")
    pbp_loader = DataNbaPbpLoader("0021900001", source_loader)
    print(pbp_loader.items[0].data)  # prints dict with the first event of the game
"""
from pbpstats.data_loader.data_nba.base import DataNbaLoaderBase
from pbpstats.resources.pbp.data_nba_pbp_item import DataNbaPbpItem


class DataNbaPbpLoader(DataNbaLoaderBase):
    """
    Loads data.nba.com source pbp data for game.
    Events are stored in items attribute as :obj:`~pbpstats.resources.pbp.data_nba_pbp_item.DataNbaPbpItem` objects

    :param str game_id: NBA Stats Game Id
    :param source_loader: :obj:`~pbpstats.data_loader.data_nba.pbp.file.DataNbaPbpFileLoader` or :obj:`~pbpstats.data_loader.data_nba.pbp.web.DataNbaPbpWebLoader` object
    """

    data_provider = "data_nba"
    resource = "Pbp"
    parent_object = "Game"

    def __init__(self, game_id, source_loader):
        self.game_id = game_id
        self.source_data = source_loader.load_data(self.game_id)
        self._make_pbp_items()

    def _make_pbp_items(self):
        self.items = [
            DataNbaPbpItem(event, item["p"])
            for item in self.data
            for event in item["pla"]
        ]

    @property
    def data(self):
        """
        returns raw JSON response data
        """
        return self.source_data["g"]["pd"]



"""
``LivePbpLoader`` loads pbp data for a game and creates :obj:`~pbpstats.resources.pbp.live_pbp_item.LivePbpItem` objects for each event

The following code will load pbp data for game id "0021900001" from a file located in a subdirectory of the /data directory

.. code-block:: python

    from pbpstats.data_loader import LivePbpFileLoader, LivePbpLoader

    source_loader = LivePbpFileLoader("/data")
    pbp_loader = LivePbpLoader("0021900001", source_loader)
    print(pbp_loader.items[0].data)  # prints dict with the first event of the game
"""
from pbpstats.data_loader.live.base import LiveLoaderBase
from pbpstats.resources.pbp.live_pbp_item import LivePbpItem


class LivePbpLoader(LiveLoaderBase):
    """
    Loads live data source pbp data for game.
    Events are stored in items attribute as :obj:`~pbpstats.resources.pbp.live_pbp_item.LivePbpItem` objects

    :param str game_id: NBA Stats Game Id
    :param source_loader: :obj:`~pbpstats.data_loader.live.pbp.file.LivePbpFileLoader` or :obj:`~pbpstats.data_loader.live.pbp.web.LivePbpWebLoader` object
    """

    data_provider = "live"
    resource = "Pbp"
    parent_object = "Game"

    def __init__(self, game_id, source_loader):
        self.game_id = game_id
        self.source_data = source_loader.load_data(self.game_id)
        self._make_pbp_items()

    def _make_pbp_items(self):
        self.items = [LivePbpItem(event) for event in self.data]

    @property
    def data(self):
        """
        returns raw JSON response data
        """
        return self.source_data["game"]["actions"]



from pbpstats.resources.enhanced_pbp.live.enhanced_pbp_item import LiveEnhancedPbpItem
from pbpstats.resources.enhanced_pbp import EndOfPeriod


class LiveEndOfPeriod(EndOfPeriod, LiveEnhancedPbpItem):
    """
    Class for end of period events
    """

    action_type = "period"
    sub_type = "end"

    def __init__(self, *args):
        super().__init__(*args)



import abc

import pbpstats
from pbpstats.resources.enhanced_pbp import Substitution


class Turnover(object):
    """
    Class for Turnover events
    """

    @property
    def is_steal(self):
        return hasattr(self, "player3_id")

    @abc.abstractclassmethod
    def is_no_turnover(self):
        pass

    @abc.abstractclassmethod
    def is_bad_pass(self):
        pass

    @abc.abstractclassmethod
    def is_lost_ball(self):
        pass

    @abc.abstractclassmethod
    def is_travel(self):
        pass

    @abc.abstractclassmethod
    def is_3_second_violation(self):
        pass

    @abc.abstractclassmethod
    def is_shot_clock_violation(self):
        pass

    @abc.abstractclassmethod
    def is_offensive_goaltending(self):
        pass

    @abc.abstractclassmethod
    def is_lane_violation(self):
        pass

    @abc.abstractclassmethod
    def is_kicked_ball(self):
        pass

    @abc.abstractclassmethod
    def is_step_out_of_bounds(self):
        pass

    @abc.abstractclassmethod
    def is_lost_ball_out_of_bounds(self):
        pass

    @abc.abstractclassmethod
    def is_bad_pass_out_of_bounds(self):
        pass

    @property
    def event_stats(self):
        """
        returns list of dicts with all stats for event
        """
        stats = []
        second_chance_stats = []
        if not self.is_no_turnover:
            team_ids = list(self.current_players.keys())
            opponent_team_id = (
                team_ids[0] if self.team_id == team_ids[1] else team_ids[1]
            )
            lineup_ids = self.lineup_ids
            if self.is_steal:
                turnover_key = (
                    pbpstats.LOST_BALL_TURNOVER_STRING
                    if self.is_lost_ball
                    else pbpstats.BAD_PASS_TURNOVER_STRING
                )
                stats.append(
                    {
                        "player_id": self.player1_id,
                        "team_id": self.team_id,
                        "stat_key": turnover_key,
                        "stat_value": 1,
                    }
                )
                steal_key = (
                    pbpstats.LOST_BALL_STEAL_STRING
                    if self.is_lost_ball
                    else pbpstats.BAD_PASS_STEAL_STRING
                )
                stats.append(
                    {
                        "player_id": self.player3_id,
                        "team_id": opponent_team_id,
                        "stat_key": steal_key,
                        "stat_value": 1,
                    }
                )
            else:
                stats.append(
                    {
                        "player_id": self.player1_id,
                        "team_id": self.team_id,
                        "stat_key": pbpstats.DEADBALL_TURNOVERS_STRING,
                        "stat_value": 1,
                    }
                )
                if self.is_travel:
                    stats.append(
                        {
                            "player_id": self.player1_id,
                            "team_id": self.team_id,
                            "stat_key": pbpstats.TRAVELS_STRING,
                            "stat_value": 1,
                        }
                    )
                elif self.is_3_second_violation:
                    stats.append(
                        {
                            "player_id": self.player1_id,
                            "team_id": self.team_id,
                            "stat_key": pbpstats.THREE_SECOND_VIOLATION_TURNOVER_STRING,
                            "stat_value": 1,
                        }
                    )
                elif self.is_step_out_of_bounds:
                    stats.append(
                        {
                            "player_id": self.player1_id,
                            "team_id": self.team_id,
                            "stat_key": pbpstats.STEP_OUT_OF_BOUNDS_TURNOVER_STRING,
                            "stat_value": 1,
                        }
                    )
                elif self.is_offensive_goaltending:
                    stats.append(
                        {
                            "player_id": self.player1_id,
                            "team_id": self.team_id,
                            "stat_key": pbpstats.OFFENSIVE_GOALTENDING_STRING,
                            "stat_value": 1,
                        }
                    )
                elif self.is_lost_ball_out_of_bounds:
                    stats.append(
                        {
                            "player_id": self.player1_id,
                            "team_id": self.team_id,
                            "stat_key": pbpstats.LOST_BALL_OUT_OF_BOUNDS_TURNOVER_STRING,
                            "stat_value": 1,
                        }
                    )
                elif self.is_bad_pass_out_of_bounds:
                    stats.append(
                        {
                            "player_id": self.player1_id,
                            "team_id": self.team_id,
                            "stat_key": pbpstats.BAD_PASS_OUT_OF_BOUNDS_TURNOVER_STRING,
                            "stat_value": 1,
                        }
                    )
                elif self.is_shot_clock_violation:
                    stats.append(
                        {
                            "player_id": self.player1_id,
                            "team_id": self.team_id,
                            "stat_key": pbpstats.SHOT_CLOCK_VIOLATION_TURNOVER_STRING,
                            "stat_value": 1,
                        }
                    )

                # sometimes events are out of order and turnover happens after sub
                # if player who turned ball over is subbed out before turnover event, fix lineup id
                events_to_check = self.get_all_events_at_current_time()
                if self.player1_id != 0 and str(self.player1_id) not in lineup_ids[
                    self.team_id
                ].split("-"):
                    # sub in wrong order - fix lineup
                    for event in events_to_check:
                        if isinstance(event, Substitution):
                            if event.outgoing_player_id == self.player1_id:
                                fixed_lineup_id = lineup_ids[self.team_id].replace(
                                    str(event.incoming_player_id),
                                    str(event.outgoing_player_id),
                                )
                                lineup_ids[self.team_id] = fixed_lineup_id

            for stat in stats:
                opponent_team_id = (
                    team_ids[0] if stat["team_id"] == team_ids[1] else team_ids[1]
                )
                stat["lineup_id"] = lineup_ids[stat["team_id"]]
                stat["opponent_team_id"] = opponent_team_id
                stat["opponent_lineup_id"] = lineup_ids[opponent_team_id]
                if self.is_second_chance_event():
                    second_chance_stats.append(
                        {
                            key: value
                            if key != "stat_key"
                            else f"{pbpstats.SECOND_CHANCE_STRING}{value}"
                            for key, value in stat.items()
                        }
                    )
                if self.is_penalty_event():
                    second_chance_stats.append(
                        {
                            key: value
                            if key != "stat_key"
                            else f"{pbpstats.PENALTY_STRING}{value}"
                            for key, value in stat.items()
                        }
                    )

        return self.base_stats + stats + second_chance_stats



"""
``DataNbaScheduleLoader`` loads schedule data for a season and
creates :obj:`~pbpstats.resources.games.data_nba_game_item.DataNbaGameItem` objects for each game

The following code will load schedule data for 2019-20 NBA Regular Season

.. code-block:: python

    from pbpstats.data_loader import DataNbaScheduleFileLoader, DataNbaScheduleLoader

    source_loader = DataNbaScheduleFileLoader("/data")
    schedule_loader = DataNbaScheduleLoader("nba", "2019-20", "Regular Season", source_loader)
    print(schedule_loader.items[0].data)  # prints dict with the first game of the season
"""
from pbpstats import (
    PLAYOFFS_STRING,
    REGULAR_SEASON_STRING,
    PLAY_IN_STRING,
)
from pbpstats.data_loader.data_nba.base import DataNbaLoaderBase
from pbpstats.resources.games.data_nba_game_item import DataNbaGameItem


class DataNbaScheduleLoader(DataNbaLoaderBase):
    """
    Loads data.nba.com source schedule data for season.
    Games are stored in items attribute
    as :obj:`~pbpstats.resources.games.data_nba_game_item.DataNbaGameItem` objects

    :param str league: Options are 'nba', 'wnba' or 'gleague'
    :param str season: Can be formatted as either 2019-20 or 2019.
    :param str season_type: Options are 'Regular Season' or 'Playoffs' or 'Play In'
    :param source_loader: :obj:`~pbpstats.data_loader.data_nba.pbp.file.DataNbaScheduleFileLoader` or :obj:`~pbpstats.data_loader.data_nba.pbp.web.DataNbaScheduleWebLoader` object
    """

    data_provider = "data_nba"
    resource = "Games"
    parent_object = "Season"

    def __init__(self, league, season, season_type, source_loader):
        self.league_string = league
        self.season_year = season.split("-")[0]
        self.source_data = source_loader.load_data(league, season)
        self._make_game_items(season_type)

    def _make_game_items(self, season_type):
        self.season_type = season_type
        self.items = [
            DataNbaGameItem(game)
            for item in self.data
            for game in item["mscd"]["g"]
            if self._is_season_type(game)
        ]

    def _is_season_type(self, game):
        if game["gid"][2] == "4" and self.season_type == PLAYOFFS_STRING:
            return True
        elif game["gid"][2] == "2" and self.season_type == REGULAR_SEASON_STRING:
            return True
        elif game["gid"][2] == "5" and self.season_type == PLAY_IN_STRING:
            return True
        return False

    @property
    def data(self):
        """
        returns raw JSON response data
        """
        return self.source_data["lscd"]



from pbpstats.resources.enhanced_pbp.live.enhanced_pbp_item import LiveEnhancedPbpItem
from pbpstats.resources.enhanced_pbp import Replay


class LiveReplay(Replay, LiveEnhancedPbpItem):
    """
    Class for replay events
    """

    action_type = "replay"

    def __init__(self, *args):
        super().__init__(*args)

    @property
    def support_ruling(self):
        return False

    @property
    def overturn_ruling(self):
        return False

    @property
    def ruling_stands(self):
        return False



from pbpstats.resources.enhanced_pbp.live.enhanced_pbp_item import LiveEnhancedPbpItem
from pbpstats.resources.enhanced_pbp import JumpBall


class LiveJumpBall(JumpBall, LiveEnhancedPbpItem):
    """
    Class for jump ball events
    """

    action_type = "jumpball"

    def __init__(self, *args):
        super().__init__(*args)



import json
import os

from pbpstats import NBA_STRING, G_LEAGUE_STRING
from pbpstats.data_loader.stats_nba.web_loader import StatsNbaWebLoader


class StatsNbaBoxscoreWebLoader(StatsNbaWebLoader):
    """
    A ``StatsNbaBoxscoreWebLoader`` object should be instantiated and passed into ``StatsNbaBoxscoreLoader`` when loading data directly from the NBA Stats API

    :param str file_directory: (optional, use it if you want to store the response data on disk)
        Directory in which data should be either stored.
        The specific file location will be `stats_boxscore_<game_id>.json` in the `/game_details` subdirectory.
        If not provided response data will not be saved on disk.
    """
    def __init__(self, file_directory=None):
        self.file_directory = file_directory

    def load_data(self, game_id):
        self.game_id = game_id
        league_url_part = (
            f"{G_LEAGUE_STRING}.{NBA_STRING}"
            if self.league == G_LEAGUE_STRING
            else self.league
        )
        self.base_url = (
            f"https://stats.{league_url_part}.com/stats/boxscoretraditionalv2"
        )
        self.parameters = {
            "GameId": self.game_id,
            "StartPeriod": 0,
            "EndPeriod": 10,
            "RangeType": 2,
            "StartRange": 0,
            "EndRange": 55800,
        }
        return self._load_request_data()

    def _save_data_to_file(self):
        if self.file_directory is not None and os.path.isdir(self.file_directory):
            file_path = (
                f"{self.file_directory}/game_details/stats_boxscore_{self.game_id}.json"
            )
            with open(file_path, "w") as outfile:
                json.dump(self.source_data, outfile)



from pbpstats.data_loader.data_nba.enhanced_pbp.web import DataNbaEnhancedPbpWebLoader


class DataNbaPossessionWebLoader(object):
    """
    A ``DataNbaPbpWebLoader`` object should be instantiated and passed into ``DataNbaPossessionLoader`` when loading data directly from the NBA Stats API

    :param str file_directory: (optional, use it if you want to store the response data on disk)
        Directory in which data should be either stored.
        The specific file location will be `data_<game_id>.json` in the `/pbp` subdirectory.
        If not provided response data will not be saved on disk.
    """
    def __init__(self, file_directory=None):
        self.file_directory = file_directory
        self.enhanced_pbp_source_loader = DataNbaEnhancedPbpWebLoader(file_directory)


import requests

from pbpstats import (
    NBA_STRING,
    G_LEAGUE_STRING,
    WNBA_STRING,
    NBA_GAME_ID_PREFIX,
    G_LEAGUE_GAME_ID_PREFIX,
    WNBA_GAME_ID_PREFIX,
)
from pbpstats import HEADERS, REQUEST_TIMEOUT
from pbpstats.resources.enhanced_pbp.stats_nba.enhanced_pbp_item import (
    StatsEnhancedPbpItem,
)
from pbpstats.resources.enhanced_pbp import (
    StartOfPeriod,
    InvalidNumberOfStartersException,
)


class StatsStartOfPeriod(StartOfPeriod, StatsEnhancedPbpItem):
    """
    Class for start of period events
    """

    event_type = 12

    def __init__(self, *args):
        super().__init__(*args)

    def get_period_starters(self, file_directory=None):
        """
        Gets player ids of players who started the period for each team.
        If players can't be determined from parsing pbp, will try to
        find them by making API request to stats.nba.com boxscore filtered by time.

        :param str file_directory: directory in which overrides subdirectory exists
            containing period starter overrides when period starters can't be determined
            from parsing pbp events
        :returns: dict with list of player ids for each team
            with players on the floor at start of period
        :raises: :obj:`~pbpstats.resources.enhanced_pbp.start_of_period.InvalidNumberOfStartersException`:
            If all 5 players that start the period for a team can't be determined.
        """
        try:
            return self._get_period_starters_from_period_events(file_directory)
        except InvalidNumberOfStartersException:
            return self._get_starters_from_boxscore_request()

    def _get_starters_from_boxscore_request(self):
        """
        makes request to boxscore url for time from period start to first event to get period starters
        """
        base_url = (
            f"https://stats.{self.league_url_part}.com/stats/boxscoretraditionalv2"
        )
        event = self
        while event is not None and event.seconds_remaining == self.seconds_remaining:
            event = event.next_event
        seconds_to_first_event = self.seconds_remaining - event.seconds_remaining

        if self.period == 1:
            start_range = 0
        elif self.period <= 4:
            start_range = int(7200 * (self.period - 1))
        else:
            start_range = int(28800 + 3000 * (self.period - 5))
        end_range = int(start_range + seconds_to_first_event * 10)
        params = {
            "GameId": self.game_id,
            "StartPeriod": 0,
            "EndPeriod": 0,
            "RangeType": 2,
            "StartRange": start_range,
            "EndRange": end_range,
        }
        starters_by_team = {}
        response = requests.get(
            base_url, params, headers=HEADERS, timeout=REQUEST_TIMEOUT
        )
        if response.status_code == 200:
            response_json = response.json()
        else:
            response.raise_for_status()

        headers = response_json["resultSets"][0]["headers"]
        rows = response_json["resultSets"][0]["rowSet"]
        players = [dict(zip(headers, row)) for row in rows]
        starters = sorted(
            players, key=lambda k: int(k["MIN"].split(":")[1]), reverse=True
        )

        for starter in starters[0:10]:
            team_id = starter["TEAM_ID"]
            player_id = starter["PLAYER_ID"]
            if team_id not in starters_by_team.keys():
                starters_by_team[team_id] = []
            starters_by_team[team_id].append(player_id)

        for team_id, starters in starters_by_team.items():
            if len(starters) != 5:
                raise InvalidNumberOfStartersException(
                    f"GameId: {self.game_id}, Period: {self.period}, TeamId: {team_id}, Players: {starters}"
                )

        return starters_by_team

    @property
    def league_url_part(self):
        if self.game_id[0:2] == NBA_GAME_ID_PREFIX:
            return NBA_STRING
        elif self.game_id[0:2] == G_LEAGUE_GAME_ID_PREFIX:
            return f"{G_LEAGUE_STRING}.{NBA_STRING}"
        elif self.game_id[0:2] == WNBA_GAME_ID_PREFIX:
            return WNBA_STRING



from pbpstats.resources.enhanced_pbp.stats_nba.enhanced_pbp_item import (
    StatsEnhancedPbpItem,
)
from pbpstats.resources.enhanced_pbp import Timeout


class StatsTimeout(Timeout, StatsEnhancedPbpItem):
    """
    Class for timeout events
    """

    event_type = 9

    def __init__(self, *args):
        super().__init__(*args)



class StatsNbaShot(object):
    """
    Class for shot data from stats.nba.com

    :param dict item: dict with shot data
    """

    def __init__(self, item):
        for key, value in item.items():
            setattr(self, key.lower(), value)

    @property
    def data(self):
        """
        returns shot data dict
        """
        return self.__dict__



"""
Instantiating a ``Season`` object will load all resources for the ``Season``
object that were set in the settings when the client was instantiated

The following code will instantiate the client and get all games for the
2019-20 NBA Regular Season and store the schedule response in a
``/response_data`` subdirectory

.. code-block:: python

    from pbpstats.client import Client

    settings = {
        "dir": "/response_data",
        "Games": {"source": "web", "data_provider": "data_nba"}
    }
    client = Client(settings)
    season = client.Season("nba", "2019-20", "Regular Season")
    for game in season.games.items:
        print(game)
"""
import inspect

import pbpstats.client as client


class Season(object):
    """
    Class for loading resource data from data loaders with a ``parent_object`` of ``Season``

    :param str league: Options are 'nba', 'wnba' or 'gleague'
    :param str season: Can be formatted as either 2019-20 or 2019.
    :param str season_type: Options are 'Regular Season' or 'Playoffs' or 'Play In'
    """

    def __init__(self, league, season, season_type):
        self.league = league
        self.season = season
        self.season_type = season_type
        attributes = inspect.getmembers(self, lambda a: not (inspect.isroutine(a)))
        data_loaders = [
            a for a in attributes if a[0].endswith(client.DATA_LOADER_SUFFIX)
        ]
        data_source_map = {
            a[0].replace(client.DATA_SOURCE_SUFFIX, ""): a[1]
            for a in attributes
            if a[0].endswith(client.DATA_SOURCE_SUFFIX)
        }
        for data_loader in data_loaders:
            attr_name = data_loader[0].replace(client.DATA_LOADER_SUFFIX, "")
            source_loader_cls = data_source_map[attr_name]
            source_loader = source_loader_cls(self.data_directory)
            data = data_loader[1](league, season, season_type, source_loader)
            resource_cls = getattr(self, attr_name)
            setattr(
                self,
                client.PATTERN.sub("_", attr_name).lower(),
                resource_cls(data.items),
            )



from pbpstats.data_loader.stats_nba.pbp.file import StatsNbaPbpFileLoader
from pbpstats.data_loader.stats_nba.shots.file import  StatsNbaShotsFileLoader


class StatsNbaEnhancedPbpFileLoader(StatsNbaPbpFileLoader):
    """
    A ``StatsNbaEnhancedPbpFileLoader`` object should be instantiated and passed into ``StatsNbaEnhancedPbpLoader`` when loading data from file

    :param str file_directory:
        Directory in which data should be loaded from.
        The specific file location will be `stats_<game_id>.json` in the `/pbp` subdirectory.
    """
    def __init__(self, file_directory):
        self.file_directory = file_directory
        self.shots_source_loader = StatsNbaShotsFileLoader(file_directory)



import abc

import pbpstats
from pbpstats.resources.enhanced_pbp import Foul


class FreeThrow(metaclass=abc.ABCMeta):
    """
    Class for free throw events
    """

    shot_type = pbpstats.FREE_THROW_STRING

    @abc.abstractproperty
    def is_made(self):
        """
        returns True if shot was made, False otherwise
        """
        pass

    @abc.abstractproperty
    def is_ft_1_of_1(self):
        pass

    @abc.abstractproperty
    def is_ft_1_of_2(self):
        pass

    @abc.abstractproperty
    def is_ft_2_of_2(self):
        pass

    @abc.abstractproperty
    def is_ft_1_of_3(self):
        pass

    @abc.abstractproperty
    def is_ft_2_of_3(self):
        pass

    @abc.abstractproperty
    def is_ft_3_of_3(self):
        pass

    @property
    def is_first_ft(self):
        """
        returns True if free throw is first of trip to the free throw line, False otherwise
        """
        return (
            self.is_ft_1_of_1
            or self.is_ft_1_of_2
            or self.is_ft_1_of_3
            or self.is_ft_1pt
            or self.is_ft_2pt
            or self.is_ft_3pt
        )

    @property
    def is_end_ft(self):
        """
        returns True if free throw is last of trip to the free throw line, False otherwise
        """
        return (
            self.is_ft_1_of_1
            or self.is_ft_2_of_2
            or self.is_ft_3_of_3
            or self.is_ft_1pt
            or self.is_ft_2pt
            or self.is_ft_3pt
        )

    @abc.abstractproperty
    def is_technical_ft(self):
        pass

    @abc.abstractproperty
    def is_ft_1pt(self):
        pass

    @abc.abstractproperty
    def is_ft_2pt(self):
        pass

    @abc.abstractproperty
    def is_ft_3pt(self):
        pass

    @property
    def shot_value(self):
        """
        returns shot value of a free throw
        Starting in 2019-20 season, the G-League added 2 and 3 point FTs
        """
        if self.is_ft_2pt:
            return 2
        if self.is_ft_3pt:
            return 3
        return 1

    @property
    def is_away_from_play_ft(self):
        """
        returns True if free throw is from an away from the play foul, False otherwise.
        """
        foul = self.foul_that_led_to_ft
        if (
            (
                (self.is_ft_1_of_1 or self.is_ft_1pt)
                or (self.is_ft_2_of_2 or self.is_ft_2pt)
            )
            and foul is not None
            and foul.is_away_from_play_foul
        ):
            made_shots_at_event_time = []
            fts_by_other_player_at_event_time = []
            events_at_event_time = self.get_all_events_at_current_time()
            for event in events_at_event_time:
                if hasattr(event, "is_made") and event.is_made:
                    if not isinstance(event, FreeThrow):
                        made_shots_at_event_time.append(event)
                    if (
                        isinstance(event, FreeThrow)
                        and event.player1_id != self.player1_id
                    ):
                        fts_by_other_player_at_event_time.append(event)

            if len(made_shots_at_event_time) == 0:
                # check for made free throw by other player - this is where player is fouled going for rebound on made FT
                if len(fts_by_other_player_at_event_time) == 0:
                    return True
                for ft in fts_by_other_player_at_event_time:
                    if ft.team_id != self.team_id:
                        return True
            else:
                made_shots_at_event_time = sorted(
                    made_shots_at_event_time, key=lambda k: k.order
                )
                if (made_shots_at_event_time[0].team_id == foul.team_id) and (
                    self.player1_id != made_shots_at_event_time[0].player1_id
                ):
                    # make sure player who made shot is not player who shot FT
                    return True

        return False

    @property
    def is_inbound_foul_ft(self):
        """
        returns True if free throw is from an inbound foul, False otherwise.
        """
        if self.is_ft_1_of_1 or self.is_ft_1pt:
            events_at_event_time = self.get_all_events_at_current_time()
            for event in events_at_event_time:
                if isinstance(event, Foul) and event.is_inbound_foul:
                    return True

        return False

    @property
    def foul_that_led_to_ft(self):
        """
        returns :obj:`~pbpstats.resources.enhanced_pbp.foul.Foul` object for the foul
        that resulted in the free throw
        """
        clock = self.clock
        # foul should be before FT so start by going backwards
        event = self
        while (
            event is not None
            and event.clock == clock
            and not (
                isinstance(event, Foul)
                and not event.is_technical
                and not event.is_double_technical
            )
        ):
            event = event.previous_event

        if (
            isinstance(event, Foul)
            and not event.is_technical
            and not event.is_double_technical
            and event.clock == clock
        ):
            return event

        # bug in pbp where foul is after FT
        event = self
        while (
            event is not None
            and event.clock == clock
            and not (
                isinstance(event, Foul)
                and not event.is_technical
                and not event.is_double_technical
            )
        ):
            event = event.next_event

        if (
            isinstance(event, Foul)
            and not event.is_technical
            and not event.is_double_technical
            and event.clock == clock
        ):
            return event
        return None

    @property
    def num_ft_for_trip(self):
        """
        returns number of shots for the trip to the free throw line
        """
        if "of 1" in self.description:
            return 1
        elif "of 2" in self.description:
            return 2
        elif "of 3" in self.description:
            return 3

    @property
    def free_throw_type(self):
        """
        returns string description of free throw type
        """
        if self.is_technical_ft:
            return "Technical"
        num_fts = self.num_ft_for_trip
        if num_fts is None:
            num_fts = self.shot_value

        if num_fts == 1:
            # check for shot before FT at same time as FT
            previous_event = self
            while (
                previous_event is not None
                and previous_event.clock == self.clock
                and not (
                    hasattr(previous_event, "is_made")
                    and previous_event.is_made
                    and not isinstance(previous_event, FreeThrow)
                )
            ):
                previous_event = previous_event.previous_event
            if (
                previous_event is not None
                and previous_event.clock == self.clock
                and (
                    hasattr(previous_event, "is_made")
                    and previous_event.is_made
                    and not isinstance(previous_event, FreeThrow)
                )
            ):
                and1_shot = previous_event
                if self.player1_id == and1_shot.player1_id:
                    return f"{and1_shot.shot_value}pt And 1"
                else:
                    return "1 Shot Away From Play"
            else:
                return "1 Shot Away From Play"
        foul_event = self.foul_that_led_to_ft
        if foul_event.is_shooting_foul or foul_event.is_shooting_block_foul:
            return f"{num_fts}pt Shooting Foul"
        elif foul_event.is_flagrant:
            if num_fts is None:
                # assume 2 shot flagrant if num_fts is None
                num_fts = 2
            return f"{num_fts} Shot Flagrant"
        elif foul_event.is_away_from_play_foul:
            return f"{num_fts} Shot Away From Play"
        elif foul_event.is_inbound_foul:
            return f"{num_fts} Shot Inbound Foul"
        elif foul_event.is_clear_path_foul:
            return f"{num_fts} Shot Clear Path"
        elif num_fts == 3:
            return "3pt Shooting Foul"
        else:
            return "Penalty"

    @property
    def event_for_efficiency_stats(self):
        """
        returns :obj:`~pbpstats.resources.enhanced_pbp.foul.Foul` object for the foul
        that resulted in the free throw. Plus/minus points should go to
        the players on the floor at the time of the foul, not the free throw.
        """
        clock = self.clock
        # foul should be before FT so start by going backwards
        event = self
        while (
            event is not None and event.clock == clock and not isinstance(event, Foul)
        ):
            event = event.previous_event

        if isinstance(event, Foul) and event.clock == clock:
            return event

        # bug in pbp where foul is after FT
        event = self
        while (
            event is not None and event.clock == clock and not isinstance(event, Foul)
        ):
            event = event.next_event

        if isinstance(event, Foul) and event.clock == clock:
            return event
        return self

    @property
    def event_stats(self):
        """
        returns list of dicts with all stats for event
        """
        stats = []
        team_ids = list(self.current_players.keys())
        is_penalty_event = self.is_penalty_event()
        is_second_chance_event = self.is_second_chance_event()
        lineup_ids = self.event_for_efficiency_stats.lineup_ids
        if self.is_made:
            stats += self._get_made_ft_stat_items(
                is_second_chance_event, is_penalty_event
            )
            # add plus minus and opponent points - used for lineup/wowy stats to get net rating
            points = self.shot_value
            for (
                team_id,
                players,
            ) in self.event_for_efficiency_stats.current_players.items():
                multiplier = 1 if team_id == self.team_id else -1
                opponent_team_id = (
                    team_ids[0] if team_id == team_ids[1] else team_ids[1]
                )
                for player_id in players:
                    stat_item = {
                        "player_id": player_id,
                        "team_id": team_id,
                        "opponent_team_id": opponent_team_id,
                        "lineup_id": lineup_ids[team_id],
                        "opponent_lineup_id": lineup_ids[opponent_team_id],
                        "stat_key": pbpstats.PLUS_MINUS_STRING,
                        "stat_value": points * multiplier,
                    }
                    stats.append(stat_item)
                    if multiplier == -1:
                        opponent_points_stat_item = {
                            "player_id": player_id,
                            "team_id": team_id,
                            "opponent_team_id": opponent_team_id,
                            "lineup_id": lineup_ids[team_id],
                            "opponent_lineup_id": lineup_ids[opponent_team_id],
                            "stat_key": pbpstats.OPPONENT_POINTS,
                            "stat_value": points,
                        }
                        stats.append(opponent_points_stat_item)
        if self.is_first_ft or self.is_technical_ft:
            stats += self._get_ft_trip_stat_items(
                is_second_chance_event, is_penalty_event
            )
        if not self.is_made:
            stats += self._get_missed_ft_stat_items(
                is_second_chance_event, is_penalty_event
            )

        opponent_team_id = team_ids[0] if self.team_id == team_ids[1] else team_ids[1]
        for stat in stats:
            if "lineup_id" not in stat.keys():
                opponent_team_id = (
                    team_ids[0] if stat["team_id"] == team_ids[1] else team_ids[1]
                )
                stat["lineup_id"] = lineup_ids[stat["team_id"]]
                stat["opponent_team_id"] = opponent_team_id
                stat["opponent_lineup_id"] = lineup_ids[opponent_team_id]

        return self.base_stats + stats

    def _get_made_ft_stat_items(self, is_second_chance_event, is_penalty_event):
        stats = []
        if self.is_ft_3pt:
            free_throw_key = pbpstats.FT_3_PT_MADE_STRING
        elif self.is_ft_2pt:
            free_throw_key = pbpstats.FT_2_PT_MADE_STRING
        elif self.is_ft_1pt:
            free_throw_key = pbpstats.FT_1_PT_MADE_STRING
        elif self.is_technical_ft:
            free_throw_key = pbpstats.TECHNICAL_FTS_MADE_STRING
        else:
            free_throw_key = pbpstats.FTS_MADE_STRING

        stats.append(
            {
                "player_id": self.player1_id,
                "team_id": self.team_id,
                "stat_key": free_throw_key,
                "stat_value": 1,
            }
        )
        if is_second_chance_event:
            second_chance_stat_key = f"{pbpstats.SECOND_CHANCE_STRING}{free_throw_key}"
            stats.append(
                {
                    "player_id": self.player1_id,
                    "team_id": self.team_id,
                    "stat_key": second_chance_stat_key,
                    "stat_value": 1,
                }
            )
        if is_penalty_event:
            penalty_stat_key = f"{pbpstats.PENALTY_STRING}{free_throw_key}"
            stats.append(
                {
                    "player_id": self.player1_id,
                    "team_id": self.team_id,
                    "stat_key": penalty_stat_key,
                    "stat_value": 1,
                }
            )
            foul_event = self.foul_that_led_to_ft
            if (
                foul_event is not None
                and foul_event.is_personal_take_foul
                and self.seconds_remaining < 60
                and self.period >= 4
            ):
                final_minute_take_foul_stat_key = (
                    f"{pbpstats.FINAL_MINUTE_PENALTY_TAKE_FOUL_STRING}{free_throw_key}"
                )
                stats.append(
                    {
                        "player_id": self.player1_id,
                        "team_id": self.team_id,
                        "stat_key": final_minute_take_foul_stat_key,
                        "stat_value": 1,
                    }
                )
        return stats

    def _get_missed_ft_stat_items(self, is_second_chance_event, is_penalty_event):
        stats = []
        if self.is_ft_3pt:
            free_throw_key = pbpstats.FT_3_PT_MISSED_STRING
        elif self.is_ft_2pt:
            free_throw_key = pbpstats.FT_2_PT_MISSED_STRING
        elif self.is_ft_1pt:
            free_throw_key = pbpstats.FT_1_PT_MISSED_STRING
        else:
            free_throw_key = pbpstats.FTS_MISSED_STRING
        stats.append(
            {
                "player_id": self.player1_id,
                "team_id": self.team_id,
                "stat_key": free_throw_key,
                "stat_value": 1,
            }
        )
        if is_second_chance_event:
            second_chance_stat_key = f"{pbpstats.SECOND_CHANCE_STRING}{free_throw_key}"
            stats.append(
                {
                    "player_id": self.player1_id,
                    "team_id": self.team_id,
                    "stat_key": second_chance_stat_key,
                    "stat_value": 1,
                }
            )
        if is_penalty_event:
            penalty_stat_key = f"{pbpstats.PENALTY_STRING}{free_throw_key}"
            stats.append(
                {
                    "player_id": self.player1_id,
                    "team_id": self.team_id,
                    "stat_key": penalty_stat_key,
                    "stat_value": 1,
                }
            )
            foul_event = self.foul_that_led_to_ft
            if (
                foul_event is not None
                and foul_event.is_personal_take_foul
                and self.seconds_remaining < 60
                and self.period >= 4
            ):
                final_minute_take_foul_stat_key = (
                    f"{pbpstats.FINAL_MINUTE_PENALTY_TAKE_FOUL_STRING}{free_throw_key}"
                )
                stats.append(
                    {
                        "player_id": self.player1_id,
                        "team_id": self.team_id,
                        "stat_key": final_minute_take_foul_stat_key,
                        "stat_value": 1,
                    }
                )
        return stats

    def _get_ft_trip_stat_items(self, is_second_chance_event, is_penalty_event):
        stats = []
        free_throw_trip_key = self.free_throw_type + " Free Throw Trips"
        stats.append(
            {
                "player_id": self.player1_id,
                "team_id": self.team_id,
                "stat_key": free_throw_trip_key,
                "stat_value": 1,
            }
        )
        if is_second_chance_event:
            second_chance_stat_key = (
                f"{pbpstats.SECOND_CHANCE_STRING}{free_throw_trip_key}"
            )
            stats.append(
                {
                    "player_id": self.player1_id,
                    "team_id": self.team_id,
                    "stat_key": second_chance_stat_key,
                    "stat_value": 1,
                }
            )
        if is_penalty_event:
            penalty_stat_key = f"{pbpstats.PENALTY_STRING}{free_throw_trip_key}"
            stats.append(
                {
                    "player_id": self.player1_id,
                    "team_id": self.team_id,
                    "stat_key": penalty_stat_key,
                    "stat_value": 1,
                }
            )
            foul_event = self.foul_that_led_to_ft
            if (
                foul_event is not None
                and foul_event.is_personal_take_foul
                and self.seconds_remaining < 60
                and self.period >= 4
            ):
                current_players = self.event_for_efficiency_stats.current_players
                offense_team_id = self.get_offense_team_id()
                for team_id, players in current_players.items():
                    final_minute_take_foul_possessions_stat_key = (
                        f"{pbpstats.FINAL_MINUTE_PENALTY_TAKE_FOUL_STRING}{pbpstats.OFFENSIVE_POSSESSION_STRING}"
                        if team_id == offense_team_id
                        else f"{pbpstats.FINAL_MINUTE_PENALTY_TAKE_FOUL_STRING}{pbpstats.DEFENSIVE_POSSESSION_STRING}"
                    )
                    for player_id in players:
                        stat_item = {
                            "player_id": player_id,
                            "team_id": team_id,
                            "stat_key": final_minute_take_foul_possessions_stat_key,
                            "stat_value": 1,
                        }
                        stats.append(stat_item)
        return stats
